unit FBClass;

interface

uses
  Classes,
  IBDatabase, IBQuery;

type
  TFBClass = class
  protected
    class function Check(DataBase: TIBDatabase): string;
  public
    class function TableExist(DataBase: TIBDatabase; TableName: string): Boolean;

    class procedure GetTables(DataBase: TIBDatabase; TablesName: TStringList);
    class procedure GetViews(DataBase: TIBDatabase; ViewsName: TStringList);
    class procedure GetTableFields(DataBase: TIBDatabase; TableName: string; FieldsName: TStringList); overload;
    class procedure GetTableFields(DataBase: TIBDatabase; TableName: string; FieldsName, FieldsType: TStringList); overload;
    class procedure GetPrimaryKeyFields(DataBase: TIBDatabase; TableName: string; FieldsName: TStringList); overload;
    class procedure GetPrimaryKeyFields(DataBase: TIBDatabase; TableName: string; FieldsName, FieldsType: TStringList); overload;

    class procedure GetRoles(DataBase: TIBDatabase; Roles: TStringList);
  end;

implementation

uses
  SysUtils;

const
  ErrTablaNotExists = 'La tabla "%s" no existe en la base de datos';

{ TFBClass }

class function TFBClass.Check(DataBase: TIBDatabase): string;
const
  TransNotAssigned = 'Transaccin no asignada';
  ErrOpenDataBase = 'Se ha producido un error al conectar con la base de datos. El mensaje de error es:'#13#13'%s';
begin
  Result := '';

  if not Assigned(DataBase.DefaultTransaction) then Result := TransNotAssigned;

  if not DataBase.Connected then
    try
      DataBase.Open;
    except
      on E: Exception do Result := Format(ErrOpenDataBase, [E.Message]);
    end;
end;

class procedure TFBClass.GetPrimaryKeyFields(DataBase: TIBDatabase;
  TableName: string; FieldsName: TStringList);
const
  SQL =
    'select rdb$field_name ' +
    'from rdb$index_segments ' +
    'where rdb$index_name = ( ' +
    '                        select rdb$index_name ' +
    '                        from rdb$relation_constraints ' +
    '                        where upper(rdb$relation_name) = upper(''%s'') and ' +
    '                              rdb$constraint_type = ''PRIMARY KEY'' ' +
    '                       ) ' +
    'order by rdb$field_position';
var
  Q: TIBQuery;
  Tmp: string;
begin
  Tmp := Check(DataBase);
  if Tmp <> '' then raise Exception.Create(Tmp);

  FieldsName.Clear;
  if not TFBClass.TableExist(DataBase, TableName) then
    raise Exception.Create(Format(ErrTablaNotExists, [TableName]));

  Q := TIBQuery.Create(nil);
  try
    Q.Database := DataBase;
    Q.Transaction := DataBase.DefaultTransaction;
    Q.SQL.Text := Format(SQL, [TableName]);
    Q.Open;
    if Q.IsEmpty then // retornem tots els camps
      TFBClass.GetTableFields(DataBase, TableName, FieldsName)
    else // retornem els camps clau
    begin
      while not Q.Eof do
      begin
        FieldsName.Add(Q.Fields[0].AsString);

        Q.Next;
      end;
    end;
  finally
    FreeAndNil(Q);
  end;
end;

class procedure TFBClass.GetPrimaryKeyFields(DataBase: TIBDatabase;
  TableName: string; FieldsName, FieldsType: TStringList);
const
  SQL =
    'select s.rdb$field_name, ' +
    '  case f.rdb$field_type ' +
    '    when 7 then ' +
    '      case f.rdb$field_sub_type ' +
    '        when 0 then ''SMALLINT'' ' +
    '        when 1 then ''NUMERIC('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '        when 2 then ''DECIMAL('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '      end ' +
    '    when 8 then ' +
    '      case f.rdb$field_sub_type ' +
    '        when 0 then ''INTEGER'' ' +
    '        when 1 then ''NUMERIC(''  || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '        when 2 then ''DECIMAL('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '      end ' +
    '    when 9 then ''QUAD'' ' +
    '    when 10 then ''FLOAT'' ' +
    '    when 12 then ''DATE'' ' +
    '    when 13 then ''TIME'' ' +
    '    when 14 then ''CHAR('' || (trunc(f.rdb$field_length / ch.rdb$bytes_per_character)) || '')'' ' +
    '    when 16 then ' +
    '      case f.rdb$field_sub_type ' +
    '        when 0 then ''BIGINT'' ' +
    '        when 1 then ''NUMERIC('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '        when 2 then ''DECIMAL('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '      end ' +
    '    when 27 then ''DOUBLE'' ' +
    '    when 35 then ''TIMESTAMP'' ' +
    '    when 37 then ''VARCHAR('' || (trunc(f.rdb$field_length / ch.rdb$bytes_per_character)) || '')'' ' +
    '    when 40 then ''CSTRING'' || (trunc(f.rdb$field_length / ch.rdb$bytes_per_character)) || '')'' ' +
    '    when 45 then ''BLOB_ID'' ' +
    '    when 261 then ''BLOB SUB_TYPE '' || f.rdb$field_sub_type ' +
    '    else ''RDB$FIELD_TYPE: '' || f.rdb$field_type || ''?'' ' +
    '  end field_type ' +
    'from rdb$index_segments s ' +
    '  join rdb$relation_fields r on r.rdb$field_name = s.rdb$field_name ' +
    '  join rdb$fields f on f.rdb$field_name = r.rdb$field_source ' +
    '  left outer join rdb$character_sets ch on (ch.rdb$character_set_id = f.rdb$character_set_id) ' +
    'where s.rdb$index_name = ( ' +
    '                          select rdb$index_name ' +
    '                          from rdb$relation_constraints ' +
    '                          where upper(rdb$relation_name) = upper(''%s'') and ' +
    '                                rdb$constraint_type = ''PRIMARY KEY'' ' +
    '                         ) and ' +
    '  upper(r.rdb$relation_name) = upper(''%s'') ' +
    'order by s.rdb$field_position';
var
  Q: TIBQuery;
  Tmp: string;
begin
  Tmp := Check(DataBase);
  if Tmp <> '' then raise Exception.Create(Tmp);

  FieldsName.Clear;
  if not TFBClass.TableExist(DataBase, TableName) then
    raise Exception.Create(Format(ErrTablaNotExists, [TableName]));

  Q := TIBQuery.Create(nil);
  try
    Q.Database := DataBase;
    Q.Transaction := DataBase.DefaultTransaction;
    Q.SQL.Text := Format(SQL, [TableName, TableName]);
    Q.Open;
    if Q.IsEmpty then // retornem tots els camps
      TFBClass.GetTableFields(DataBase, TableName, FieldsName, FieldsType)
    else // retornem els camps clau
    begin
      while not Q.Eof do
      begin
        FieldsName.Add(Q.Fields[0].AsString);
        FieldsType.Add(Q.Fields[1].AsString);

        Q.Next;
      end;
    end;
  finally
    FreeAndNil(Q);
  end;
end;

class procedure TFBClass.GetRoles(DataBase: TIBDatabase;
  Roles: TStringList);
const
  SQL = 'select rdb$role_name, rdb$owner_name from rdb$roles';
var
  Q: TIBQuery;
  Tmp: string;
begin
  Tmp := Check(DataBase);
  if Tmp <> '' then raise Exception.Create(Tmp);

  Roles.Clear;
  Q := TIBQuery.Create(nil);
  try
    Q.Database := DataBase;
    Q.Transaction := DataBase.DefaultTransaction;
    Q.SQL.Text := SQL;
    Q.Open;
    while not Q.Eof do
    begin
      Roles.Add(Q.Fields[0].AsString + ';' + Q.Fields[1].AsString);
      Q.Next;
    end;
  finally
    FreeAndNil(Q);
  end;
end;

class procedure TFBClass.GetTableFields(DataBase: TIBDatabase;
  TableName: string; FieldsName: TStringList);
const
  SQL =
    'select rdb$field_name ' +
    'from rdb$relation_fields ' +
    'where upper(rdb$relation_name) = upper(''%s'') and rdb$system_flag = 0 ' +
    'order by rdb$field_position';
var
  Q: TIBQuery;
  Tmp: string;
begin
  Tmp := Check(DataBase);
  if Tmp <> '' then raise Exception.Create(Tmp);

  FieldsName.Clear;
  Q := TIBQuery.Create(nil);
  try
    Q.Database := DataBase;
    Q.Transaction := DataBase.DefaultTransaction;
    Q.SQL.Text := Format(SQL, [TableName]);
    Q.Open;
    while not Q.Eof do
    begin
      FieldsName.Add(Q.Fields[0].AsString);
      Q.Next;
    end;
  finally
    FreeAndNil(Q);
  end;
end;

class procedure TFBClass.GetTableFields(DataBase: TIBDatabase;
  TableName: string; FieldsName, FieldsType: TStringList);
const
  SQL =
    'select r.rdb$field_name, ' +
    '  case f.rdb$field_type ' +
    '    when 7 then ' +
    '      case f.rdb$field_sub_type ' +
    '        when 0 then ''SMALLINT'' ' +
    '        when 1 then ''NUMERIC('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '        when 2 then ''DECIMAL('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '      end ' +
    '    when 8 then ' +
    '      case f.rdb$field_sub_type ' +
    '        when 0 then ''INTEGER'' ' +
    '        when 1 then ''NUMERIC(''  || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '        when 2 then ''DECIMAL('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '      end ' +
    '    when 9 then ''QUAD'' ' +
    '    when 10 then ''FLOAT'' ' +
    '    when 12 then ''DATE'' ' +
    '    when 13 then ''TIME'' ' +
    '    when 14 then ''CHAR('' || (trunc(f.rdb$field_length / ch.rdb$bytes_per_character)) || '')'' ' +
    '    when 16 then ' +
    '      case f.rdb$field_sub_type ' +
    '        when 0 then ''BIGINT'' ' +
    '        when 1 then ''NUMERIC('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '        when 2 then ''DECIMAL('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '      end ' +
    '    when 27 then ''DOUBLE'' ' +
    '    when 35 then ''TIMESTAMP'' ' +
    '    when 37 then ''VARCHAR('' || (trunc(f.rdb$field_length / ch.rdb$bytes_per_character)) || '')'' ' +
    '    when 40 then ''CSTRING'' || (trunc(f.rdb$field_length / ch.rdb$bytes_per_character)) || '')'' ' +
    '    when 45 then ''BLOB_ID'' ' +
    '    when 261 then ''BLOB SUB_TYPE '' || f.rdb$field_sub_type ' +
    '    else ''RDB$FIELD_TYPE: '' || f.rdb$field_type || ''?'' ' +
    '  end field_type ' +
    'from rdb$relation_fields r ' +
    '  join rdb$fields f on f.rdb$field_name = r.rdb$field_source ' +
    '  left outer join rdb$character_sets ch on (ch.rdb$character_set_id = f.rdb$character_set_id) ' +
    'where upper(r.rdb$relation_name) = upper(''%s'') ' +
    'order by r.rdb$field_position';
var
  Q: TIBQuery;
  Tmp: string;
begin
  Tmp := Check(DataBase);
  if Tmp <> '' then raise Exception.Create(Tmp);

  FieldsName.Clear;
  Q := TIBQuery.Create(nil);
  try
    Q.Database := DataBase;
    Q.Transaction := DataBase.DefaultTransaction;
    Q.SQL.Text := Format(SQL, [TableName]);
    Q.Open;
    while not Q.Eof do
    begin
      FieldsName.Add(Q.Fields[0].AsString);
      FieldsType.Add(Q.Fields[1].AsString);

      Q.Next;
    end;
  finally
    FreeAndNil(Q);
  end;
end;

class procedure TFBClass.GetTables(DataBase: TIBDatabase;
  TablesName: TStringList);
const
  SQL =
    'select rdb$relation_name ' +
    'from rdb$relations ' +
    'where rdb$view_blr is null and (rdb$system_flag is null or rdb$system_flag = 0)';
var
  Q: TIBQuery;
  Tmp: string;
begin
  Tmp := Check(DataBase);
  if Tmp <> '' then raise Exception.Create(Tmp);

  TablesName.Clear;
  Q := TIBQuery.Create(nil);
  try
    Q.Database := DataBase;
    Q.Transaction := DataBase.DefaultTransaction;
    Q.SQL.Text := SQL;
    Q.Open;
    while not Q.Eof do
    begin
      TablesName.Add(Q.Fields[0].AsString);
      Q.Next;
    end;
  finally
    FreeAndNil(Q);
  end;
end;

class procedure TFBClass.GetViews(DataBase: TIBDatabase;
  ViewsName: TStringList);
const
  SQL =
    'select rdb$relation_name ' +
    'from rdb$relations ' +
    'where rdb$view_blr is not null and (rdb$system_flag is null or rdb$system_flag = 0)';
var
  Q: TIBQuery;
  Tmp: string;
begin
  Tmp := Check(DataBase);
  if Tmp <> '' then raise Exception.Create(Tmp);

  ViewsName.Clear;
  Q := TIBQuery.Create(nil);
  try
    Q.Database := DataBase;
    Q.Transaction := DataBase.DefaultTransaction;
    Q.SQL.Text := SQL;
    Q.Open;
    while not Q.Eof do
    begin
      ViewsName.Add(Q.Fields[0].AsString);
      Q.Next;
    end;
  finally
    FreeAndNil(Q);
  end;
end;

class function TFBClass.TableExist(DataBase: TIBDatabase;
  TableName: string): Boolean;
const
  SQL =
    'select count(*) ' +
    'from rdb$relations ' +
    'where rdb$view_blr is null and (rdb$system_flag is null or rdb$system_flag = 0) and upper(rdb$relation_name) = upper(''%s'')';
var
  Q: TIBQuery;
  Tmp: string;
begin
  Tmp := Check(DataBase);
  if Tmp <> '' then raise Exception.Create(Tmp);

  Q := TIBQuery.Create(nil);
  try
    Q.Database := DataBase;
    Q.Transaction := DataBase.DefaultTransaction;
    Q.SQL.Text := Format(SQL, [TableName]);
    Q.Open;
    Result := Q.Fields[0].AsInteger > 0;
  finally
    FreeAndNil(Q);
  end;
end;

end.
